using System;

public class BlackScholes_gold   {

 ///////////////////////////////////////////////////////////////////////////////
 // Polynomial approximation of cumulative normal distribution function
 ///////////////////////////////////////////////////////////////////////////////
 public static double CND(double d){
    const double       A1 = 0.31938153;
    const double       A2 = -0.356563782;
    const double       A3 = 1.781477937;
    const double       A4 = -1.821255978;
    const double       A5 = 1.330274429;
    const double RSQRT2PI = 0.39894228040143267793994605993438;

    double
        K = 1.0 / (1.0 + 0.2316419 * Math.Abs(d));

    double
        cnd = RSQRT2PI * Math.Exp(- 0.5 * d * d) *
        (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5)))));

    if(d > 0)
        cnd = 1.0 - cnd;

    return cnd;
 }


 ///////////////////////////////////////////////////////////////////////////////
 // Black-Scholes formula for both call and put
 ///////////////////////////////////////////////////////////////////////////////
 static void BlackScholesBodyCPU(
    out float callResult,
    out float putResult,
    float Sf, //Stock price
    float Xf, //Option strike
    float Tf, //Option years
    float Rf, //Riskless rate
    float Vf  //Volatility rate
 ){
    double S = Sf, X = Xf, T = Tf, R = Rf, V = Vf;

    double sqrtT = Math.Sqrt(T);
    double    d1 = (Math.Log(S / X) + (R + 0.5 * V * V) * T) / (V * sqrtT);
    double    d2 = d1 - V * sqrtT;
    double CNDD1 = CND(d1);
    double CNDD2 = CND(d2);

    //Calculate Call and Put simultaneously
    double expRT = Math.Exp(- R * T);
    callResult   = (float)(S * CNDD1 - X * expRT * CNDD2);
    putResult    = (float)(X * expRT * (1.0 - CNDD2) - S * (1.0 - CNDD1));
 }


 ////////////////////////////////////////////////////////////////////////////////
 // Process an array of optN options
 ////////////////////////////////////////////////////////////////////////////////
 public static void BlackScholesCPU(
    float[] h_CallResult,
    float[] h_PutResult,
    float[] h_StockPrice,
    float[] h_OptionStrike,
    float[] h_OptionYears,
    float Riskfree,
    float Volatility,
    int optN
 ){
    for(int opt = 0; opt < optN; opt++)
        BlackScholesBodyCPU(
            out h_CallResult[opt],
            out h_PutResult[opt],
            h_StockPrice[opt],
            h_OptionStrike[opt],
            h_OptionYears[opt],
            Riskfree,
            Volatility
        );
 }

}