//
// Copyright (C) 2009 Yury Serdyuk
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.IO;
using System.Threading;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Collections;

namespace MCSharp
{
 public class SocketStateObject {
  private Socket s;

  public SocketStateObject( Socket s ) {
   this.s = s;
  }

  public void CloseSocketCallback( IAsyncResult res ) {
   s.Close();
  }

  public void DontCloseSocketCallback( IAsyncResult res ) {
   // Just do nothing...
  }
 }

 public class TCP
 {
  public static void DeliverMovableMethod( Node node, object o, string methodName, object[] parameters )
  {
   MCStatistics statistics = MCStatistics.CurrentStatistics;
   statistics.numberOfMovableCalls++;
   if ( !Session.Distributed ||
        (node.Host.Equals( Communicator.Host ) && node.Port == Communicator.Port) )
   {
    MemoryStream ms = new MemoryStream();
    
    int ms2_size = 16384;
    
    for ( int i = 0; i < parameters.Length; i++ )
     if ( parameters[i].GetType().IsArray )
      ms2_size += GetArraySize ( parameters [ i ] );      
                
    MemoryStream ms2 = new MemoryStream(ms2_size);
    DateTime a = DateTime.Now;
    Serialization.Serialize( ms, o );
    Serialization.Serialize( ms2, parameters );
                    
    statistics.totalTimeOfMovableCallsSerialization += DateTime.Now.Subtract( a ).Ticks;
    ms.Seek( 0, SeekOrigin.Begin );
    o = Serialization.Deserialize( ms );
    ms2.Seek( 0, SeekOrigin.Begin );
    parameters = (object[])Serialization.Deserialize( ms2 );
    
    new MovableMethodHandler( new Agent( o, methodName, parameters) ).StartInNewThread();
    return;
   }
   
   TCP.SendCommandAndObject( node.Host, node.Port, "movablecall\n",
    new Agent( o, methodName, parameters ) );
   
   
  }

  public static void DeliverMovableMethod( object o, string methodName, object[] parameters )
  {
   if ( !Session.Distributed )
   {
    DeliverMovableMethod( CommWorld.Nodes [0], o, methodName, parameters );
    return;
   }

   string command = "getfreenode " + Session.Guid + "\n";

   Socket client = CreateClientSocket( Communicator.RMHost, Communicator.RMPort );
   TCP.SendMessage( client, command );
   string host = TCP.ReadLine( client );
      
   int port = int.Parse( TCP.ReadLine( client ) );
   DeliverMovableMethod( new Node( host, port ), o, methodName, parameters );
  }
  

  public static void DeliverChannelMessage( channel c, object[] parameters )
  {
   if ( Globals.IsDebug ) {
    Console.WriteLine( "TCP.DeliverChannelMessage: " +
     Communicator.Host + ":" + Communicator.Port + " -> " + c.Host + ":" + c.Port );
   }

   MCStatistics statistics = MCStatistics.CurrentStatistics;
   statistics.numberOfChannelMessages++;
   Agent agent = new Agent( c, parameters );
   if ( !Session.Distributed )
   {
    new ChannelMessageHandler( agent ).Start();
    return;
   }

   if ( c.Host.Equals( Communicator.Host ) &&
        c.Port == Communicator.Port ) {
    new ChannelMessageHandler( agent ).Start();
   }
   else {
    
    SendCommandAndObject( c.Host, c.Port, "channelmessage\n", agent );
   }
  }

  public static void SendCommandAndObject( string host, int port, string header, object o )
  {
   Socket client = CreateClientSocket( host, port );
   if ( client == null ) {
    throw new Exception( "Error: T0006! Can't create socket (" + host + ":" + port + ")!" );
   }
   SendMessage( client, header );
     
   
   int ms2_size = 16384;
   object[] parameters = ((Agent)o).Parameters;
   for ( int i = 0; i < parameters.Length; i++ )
    if ( parameters[i].GetType().IsArray )
     ms2_size += GetArraySize ( parameters [ i ] );         
             
   MemoryStream ms = new MemoryStream(ms2_size);
   DateTime  a = DateTime.Now;
   Serialization.Serialize( ms, o );
      
   if ( header.StartsWith( "movable" ) ) {
    MCStatistics.CurrentStatistics.totalTimeOfMovableCallsSerialization += DateTime.Now.Subtract( a ).Ticks;
   }
   else if ( header.StartsWith( "channel" ) ) {
    MCStatistics.CurrentStatistics.totalTimeOfChannelMessagesSerialization += DateTime.Now.Subtract( a ).Ticks;
   }

   byte[] bytes = ms.ToArray();
   //byte[] bytes = ms.GetBuffer();
   
   // Sending length of bytes. We will need this on client in order to create correct buffer
   
   SendMessage ( client, bytes.Length + "\n" ); 
   SendMessage( client, bytes, bytes.Length );
  
   client.Close();
  }

  public static void SendCommandAndObjectAsync( string host, int port, string header, object o )
  {
   Socket client = CreateClientSocket( host, port );
   if ( client == null ) {
    throw new Exception( "Error: T0006! Can't create socket (" + host + ":" + port + ")!" );
   }
   SendMessage( client, header );
   MemoryStream ms = new MemoryStream();

   DateTime a = DateTime.Now;
   Serialization.Serialize( ms, o );

   if ( header.StartsWith( "movable" ) ) {
    MCStatistics.CurrentStatistics.totalTimeOfMovableCallsSerialization += DateTime.Now.Subtract( a ).Ticks;
   }
   else if ( header.StartsWith( "channel" ) ) {
    MCStatistics.CurrentStatistics.totalTimeOfChannelMessagesSerialization += DateTime.Now.Subtract( a ).Ticks;
   }

   byte[] bytes = ms.ToArray();

   // Sending length of bytes. We will need this on client in order to create correct buffer
   SendMessage( client, bytes.Length + "\n" );

   SendMessageAsync( client, bytes, bytes.Length );
  }

  public static void SendMessage( Socket s, string message )
  {
   byte[] buf = Encoding.ASCII.GetBytes( message.ToCharArray() );
   SendMessage( s, buf, buf.Length );
  }

  public static void SendMessageAsync( Socket s, string message )
  {
   byte[] buf = Encoding.ASCII.GetBytes( message.ToCharArray() );
   SendMessageAsync( s, buf, buf.Length );
  }

  public static void SendMessage( Socket s, byte[] buf, int len )
  {
   MCStatistics statistics = MCStatistics.CurrentStatistics;
   DateTime a = DateTime.Now;

   
   int  offset = 0;
   while ( offset < len )  {

    int n = s.Send  ( buf, offset, len - offset, SocketFlags.None );
        
    offset += n;
   }

   if ( statistics != null ) { // statistics is null on RM - we don't need to collect stats for RM
    statistics.numberOfBytesSentAcrossNetwork += len;
    statistics.totalTimeOfTransportingMessages += DateTime.Now.Subtract( a ).Ticks;
   }

  }

  ////////////////////////////////////////////////////////////////////////////////////

  public static void SendMessageAsync( Socket s, byte[] buf, int len )
  {
   MCStatistics statistics = MCStatistics.CurrentStatistics;
   DateTime a = DateTime.Now;

   SocketStateObject sso = new SocketStateObject( s );
   s.BeginSend( buf, 0, buf.Length, 0, new AsyncCallback( sso.CloseSocketCallback ), null );
   /*
   // Note: We can't use cycles for asynchronous sending
   int  size;
   int  offset = 0;
   while ( offset < len )  {
    if ( offset + 2048 <= len )
     size = 2048;
    else
     size = len - offset;

    while ( !s.Poll( -1, SelectMode.SelectWrite ) );
    s.Send( buf, offset, size, SocketFlags.None );
    offset += size;
   }
   */
   if ( statistics != null ) { // statistics is null on RM - we don't need to collect stats for RM
//    statistics.numberOfBytesSentAcrossNetwork += len;
    statistics.numberOfBytesSentAcrossNetwork += buf.Length;
    statistics.totalTimeOfTransportingMessages += DateTime.Now.Subtract( a ).Ticks;
   }
  }

  public static Socket SendMessage( string host, int port, string message )
  {
   if ( Globals.IsDebug ) {
    Console.WriteLine( "-> " + host + ":" + port + ": " + message );
   }

   Socket client = CreateClientSocket( host, port );

   if ( client != null )
   {
    SendMessage( client, message );
   }
   else {
    Console.WriteLine( "Warning: client socket is null!" );
   }
   return client;
  }

  public static Socket SendMessageAsync( string host, int port, string message )
  {
   if ( Globals.IsDebug ) {
    Console.WriteLine( "-> " + host + ":" + port + ": " + message );
   }
   Socket client = CreateClientSocket( host, port );
   if ( client != null )
   {
    SendMessageAsync( client, message );
   }
   else {
    Console.WriteLine( "Warning: client socket is null!" );
   }
   return client;
  }

  ///////////////////////////////////////////////////////////////////////////

  public static void ReadDataToBuffer( Socket s, byte[] buffer   )   {

   int   offset = 0;
   int   n      = 1;

   while ( n > 0 )   {

    n = s.Receive ( buffer, offset, buffer.Length - offset, SocketFlags.None );
    offset += n;

   }

  }

  /////////////////////////////////////////////////////////////////////////////

  public static void ReadDataToStream( Socket s, MemoryStream ms )
  {
   Byte[] bytes = new Byte[ 1024 ];
   int n = 1;

   while ( n > 0 )   {

    while ( ! s.Poll ( -1, SelectMode.SelectRead ) );

    n = s.Receive( bytes, bytes.Length, SocketFlags.None );
    ms.Write ( bytes, 0, n );
   }
  }
  
 /////////////////////////////////////////////////////////////////////////////
  
     public static void ReadDataToStreamWithLength(Socket socket,
                                         MemoryStream mstream, int length)
     {   
         byte[] bytes = new byte[length];

         
         int offset = 0;
         int n;

         while ( length - offset > 0)
         {
             n = socket.Receive(bytes, offset, length - offset, SocketFlags.None);             
             offset += n;
         }
         
         
         mstream.Write ( bytes, 0, length );
     }
     
 ///////////////////////////////////////////////////////////////////////////////

  public static string ReadLine( Socket s )
  {
   byte[] buffer = new byte [1024];
   int pos = 0;
   int n;
   

   bool  continRead = true;

   bool  error      = false;

   
   while ( continRead )
   {
    //buffer [pos++] = (Byte) bytes[0];
    //if ( s.Poll( -1, SelectMode.SelectRead ) ) {
    //s.Receive(bytes, 0, 1, SocketFlags.None);

    //while ( ! s.Poll ( -1, SelectMode.SelectRead ) );

    n = s.Receive ( buffer, pos, 1, SocketFlags.None );

    if ( n != 0 )

     if ( buffer [ pos ] == 10 )
      continRead = false;
     else
      pos++;

    else {
     Console.WriteLine ( "Error ! Received " + n + " bytes ..." );
     continRead = false;
     error      = true;
    }

   }

   if ( error )   {

    if ( pos > 0 )
     Console.WriteLine ( "Reseived message:" + Encoding.ASCII.GetString ( buffer, 0, pos ) );
    return ( "" );

   }
   return Encoding.ASCII.GetString( buffer, 0, pos );
  }

  public static int MAXNUMBEROFTRIES = 200;

  public static Socket CreateClientSocket( string host, int port )
  {
   return CreateClientSocket( host, port, MAXNUMBEROFTRIES );
  }

  public static Socket CreateClientSocket( string host, int port, int numberOfTries )
  {
   IPAddress  hostadd = Dns.Resolve( host ).AddressList[0];
   IPEndPoint EPhost = new IPEndPoint( hostadd, port );
      
   int curTries = 0;
   
   Socket s = new Socket ( AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
   bool   connected = false;
   

   //while ( s == null && curTries < numberOfTries )
   while ( connected == false && curTries < numberOfTries )
   {
    try {
     //s = new Socket( AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
     s.Connect( EPhost );
     connected = s.Connected;
     //Console.WriteLine ( "curTries=" + curTries + " connected=" + connected );
     //return s;
    }
    catch ( SocketException se )
    {
     //s = null;
     curTries++;
     //Console.WriteLine ( "curTries=" + curTries );
     Thread.Sleep ( 400 );
    }
    //Thread.Sleep( 10 );
    //curTries++;
   }

   if ( connected == true )
    return ( s );
   else
    throw new Exception( "Error: T001! Can't create client socket ( " +
     host + " " + port + " )! Try to increase TCP.MAXNUMBEROFTRIES!" );

  }

  public static Socket CreateServerSocket( int from, int to, out int port ) {
   Socket serverSocket;
   Random rnd = new Random( (int)DateTime.Now.Ticks );
   int diff = to - from;
   port = -1;
   int i = 0;

   while ( i < MAXNUMBEROFTRIES ) {
    try {
     int portNumber = from + (int) Math.Floor( rnd.NextDouble() * diff );
     IPEndPoint ipep = new IPEndPoint( IPAddress.Any, portNumber );
     serverSocket = new Socket( AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
     serverSocket.Bind( ipep );
     serverSocket.Listen( 100 );
     port = portNumber;
     return serverSocket;
    }
    catch ( Exception ex )
    {Console.WriteLine ( "CreateServerSocket Exception ..." );
     i++;
    };
   }
   throw new Exception( "Error: T100! Can't bind to any port on this system! Try to increase TCP.MAXNUMBEROFTRIES!" );
  }
  


  public static void SendCommandAndArray( string host, int port, string header, byte[] array, int offset, int count )
  {
   Socket client = CreateClientSocket( host, port );
   if ( client == null ) {
    throw new Exception( "Error: T0016! Can't create socket (" + host + ":" + port + ")!" );
   }
   TCP.SendMessage( client, header );

   // Sending length of bytes. We will need this on client in order to create correct buffer
   SendMessage( client, count + "\n" );

   SendMessage( client, array, offset, count );
   client.Close();
  }

  public static void SendMessage( Socket s, byte[] buf, int offset, int count )
  {
   MCStatistics statistics = MCStatistics.CurrentStatistics;
   DateTime a = DateTime.Now;

   int size;
   while ( offset < count )  {
    if ( offset + 2048 <= count )
     size = 2048;
    else
     size = count - offset;

    while ( !s.Poll( -1, SelectMode.SelectWrite ) );
    s.Send( buf, offset, size, SocketFlags.None );
    offset += size;
   }
   if ( statistics != null ) { // statistics is null on RM - we don't need to collect stats for RM
    statistics.numberOfBytesSentAcrossNetwork += count;
    statistics.totalTimeOfTransportingMessages += DateTime.Now.Subtract( a ).Ticks;
   }
  }

    
    
  //------------------- Function to get value through the handler
  
  public static object getRemoteValue  ( handler handl, object[] args ) {
  
   if ( ! Session.Distributed || ( Communicator.Host.Equals ( handl.Host ) &&
                                   Communicator.Port == handl.Port            ) ) {
    object obj = CommExec.Objects [ handl.ObjectNumber ];
    Type   t   = obj.GetType();
    MethodInfo minfo = t.GetMethod ( handl.MethodName );

    return ( minfo.Invoke ( obj, args ) );

   }
   else   {
      
    String command = "handlerGet\n";
    MCSharp.Agent agent = new Agent ( handl, args );

    Socket socket = CreateClientSocket(handl.Host, handl.Port);
    SendMessage ( socket, command );
            
    MemoryStream mstream = new MemoryStream();
    Serialization.Serialize(mstream, agent);
    //byte[] bytes = mstream.ToArray();
    byte[] bytes = mstream.GetBuffer();
    
    SendMessage(socket, bytes.Length + "\n");
    SendMessage(socket, bytes, bytes.Length);
                
    int bufLength = int.Parse(TCP.ReadLine(socket));
    
    if ( bufLength == 0 )
     return null;
    else {            
     mstream = new MemoryStream(bufLength);
     ReadDataToStreamWithLength ( socket, mstream, bufLength );
        
     mstream.Seek ( 0, SeekOrigin.Begin );
     object result = Serialization.Deserialize(mstream);
    
     return ( result );
    }

   }
   
  }
  
  public static int GetArraySize ( object a ) {
   
   int   size  = 1;
   Type  aType = a.GetType();
       
   object _a = a;
         
   while ( aType.IsArray ) {
    size *= ((Array)_a).Length;
    int rank = ((Array)_a).Rank;
    int[] zero_ranks = new int [ rank ];
    for ( int i = 0; i < rank; i++ )
     zero_ranks [ i ] = 0;
    _a = ((Array)_a).GetValue ( zero_ranks );
    aType = _a.GetType();
   }
                                   
   Type elemType = aType;
   switch ( Type.GetTypeCode ( elemType ) )
   {
    case TypeCode.Int32:
     size *= 4;
     break;
    case TypeCode.UInt32:
     size *= 4;
     break;
    case TypeCode.Int64:
     size *= 8;
     break;
    case TypeCode.UInt64:
     size *= 8;
     break;
    case TypeCode.Single:
     size *= 4;
     break;
    case TypeCode.Double:
     size *= 8;
     break;
    default:
     break;
   }
   
   return ( size );
                                                                                                                      
  }      //  The end of GetArraySize
                                                                                                                       
 }
}
